"""
This package contains basic utilities for the backend engine,
specifically for :class:`returnn.tf.engine.Engine` or :class:`returnn.theano.engine.Engine`,
and also the base class :class:`returnn.engine.base.EngineBase`.
"""
